from typing import Dict, Any
import requests
from urllib import parse
import logging
from requests.models import Response

import annolab
from annolab import endpoints
from annolab.util.cached_property import cached_property

class ApiHelper(object):

  def __init__(
    self,
    api_key = None,
    api_url = 'https://api.annolab.ai',
  ):
    self.api_url = api_url
    self.api_key = api_key or annolab.api_key


  @property
  def __auth_header(self, key: str = None):
    key = key or self.api_key
    return { 'Authorization': f'Api-Key {key}' }


  @cached_property
  def api_key_info(self):
    return self.get_request(endpoints.ApiKey.get_api_key_info()).json()


  '''
    Returns the default group to use for the api key.
    The default group is the group representing the single user.
  '''
  @cached_property
  def default_owner(self):
    default_owner = None
    for group in self.api_key_info['groups']:
      if (group['isSingleUser'] is True):
        default_owner = group

    return default_owner


  def get_request(self, path: str, body: Dict[str, Any] = None, params: dict = None) -> Response:
    resp = requests.get(
      parse.urljoin(self.api_url, path),
      headers=self.__auth_header,
      json=body
    )

    self.__handle_non_2xx_response(resp)

    return resp


  def post_request(self, path: str, body: Dict[str, Any] = None, params: dict = None, timeout: float = 30.0):
    resp = requests.post(
      parse.urljoin(self.api_url, path),
      headers=self.__auth_header,
      json=body,
      params=params,
      timeout=timeout
    )

    self.__handle_non_2xx_response(resp)

    return resp


  def put_request(self, path: str, data: Any = None, headers = None, params: dict = None, timeout: float = 30.0):
    resp = requests.put(
      parse.urljoin(self.api_url, path),
      headers=headers,
      data=data,
      params=params,
      timeout=timeout
    )

    self.__handle_non_2xx_response(resp)

    return resp


  def __handle_non_2xx_response(self, resp: Response):
    if (resp.status_code >= 300):
      try:
        resp_body = resp.json()
      except:
        resp_body = {}

      message = resp_body['message'] if 'message' in resp_body else 'Unknown Error'

      logging.error(f'{resp.request.method} {resp.request.path_url} failed with message: {message}')
      resp.raise_for_status()



from http import HTTPStatus
from logging import Logger
import os
import re
import shutil
import tempfile
from unicodedata import category
from uuid import uuid4
from typing import Union, List

import jsonlines
from requests.exceptions import HTTPError

logger = Logger(__name__)

class ProjectImport:

  source_file: str = None
  bounds_file: str = None
  annotations_file: str = None
  layers_file: str = None
  relations_file: str = None
  atntypes_file: str = None

  # Maps original source id to source name + directory
  source_map: dict = {}
  annotation_map: dict = {}

  def __init__(
    self,
    export_filepath: str,
    project,
    groupId: Union[str, int]
  ):
    self.export_filepath = export_filepath
    self.project = project
    self.groupId = groupId
    self.unpack_target_dir = os.path.join(tempfile.gettempdir(), str(uuid4()))


  def unzip_export(self):
    if not os.path.exists(self.unpack_target_dir):
      os.mkdir(self.unpack_target_dir)

    shutil.unpack_archive(self.export_filepath, self.unpack_target_dir)
    self.__find_entity_files()


  def import_all(self):
    self.import_sources()
    self.import_annotation_types()
    self.import_layers()
    self.import_annotations()
    self.import_relations()


  def cleanup(self):
    shutil.rmtree(self.unpack_target_dir, ignore_errors=True)


  def create_source_map(self):
    filepath = os.path.join(self.unpack_target_dir, self.source_file)
    with jsonlines.open(filepath) as sources:
      for source in sources:
        self.source_map[source.get('sourceId')] = [source.get('sourceName'), source.get('directoryName')]


  def import_sources(self):
    filepath = os.path.join(self.unpack_target_dir, self.source_file)
    with jsonlines.open(filepath) as sources:
      for source in sources:
        self.create_source(source)


  def import_annotation_types(self):
    types_filepath = os.path.join(self.unpack_target_dir, self.atntypes_file)

    with jsonlines.open(types_filepath) as atn_types:
      for atn_type in atn_types:
        try:
          self.project.create_annotation_type(
            name=atn_type.get('name'),
            color=atn_type.get('color'),
            is_relation=atn_type.get('isRelation'),
            is_document_classification=atn_type.get('isDocumentClassification'),
            category=atn_type.get('category'))
        except HTTPError as e:
          if (e.response.status_code == HTTPStatus.CONFLICT):
            logger.warning(f'Annotation type {atn_type.get("name")} already exists. Skipping')
          else:
            raise e


  def import_layers(self):
    layers_filepath = os.path.join(self.unpack_target_dir, self.layers_file)
    with jsonlines.open(layers_filepath) as layers:
      for layer in layers:
        try:
          self.project.create_annotation_layer(
            name=layer.get('name'),
            is_gold=layer.get('isGoldSet'),
            description=layer.get('description')
          )
        except HTTPError as e:
          if (e.response.status_code == HTTPStatus.CONFLICT):
            logger.warning(f'Annotation type {layer.get("name")} already exists. Skipping')
          else:
            raise e


  def import_annotations(self):
    batch_size = 500
    annotations_filepath = os.path.join(self.unpack_target_dir, self.annotations_file)

    batch = []

    def insert_batch(batch: List):
      created = self.project.create_bulk_annotations(batch, dedup=True)
      for atn in created:
        self.annotation_map[str(atn.get('clientId'))] = atn

    with jsonlines.open(annotations_filepath) as annotations:
      for annotation in annotations:
        source = self.source_map.get(annotation.get('sourceId'), None)
        if (source is None):
          logger.info(f'Skipping annotation for source {annotation.get("sourceId")}, source has not been imported.')
          continue

        sourceName = source[0]
        dirName = source[1]

        batch.append({
          'type': annotation.get('typeName'),
          'value': annotation.get('value'),
          'offsets': annotation.get('offsets'),
          'text_bounds': annotation.get('textBounds'),
          'image_bounds': annotation.get('imageBounds'),
          'client_id': annotation.get('id'),
          'layer': annotation.get('layerName'),
          'page': annotation.get('pageNumber'),
          'endPage': annotation.get('endPageNumber'),
          'source': sourceName,
          'directory': dirName,
          'project': self.project.id
        })
        if (len(batch) >= batch_size):
          insert_batch(batch)
          batch=[]

    # Insert final batch
    if (len(batch) > 0):
      insert_batch(batch)


  def import_relations(self):
    batch_size = 500
    relations_filepath = os.path.join(self.unpack_target_dir, self.relations_file)
    batch = []

    with jsonlines.open(relations_filepath) as relations:
      for rln in relations:
        batch.append({
          'annotations': [
            self.annotation_map.get(str(rln.get('predecessorId'))).get('id'),
            self.annotation_map.get(str(rln.get('successorId'))).get('id')
          ],
          'type': rln.get('typeName'),
          'value': rln.get('value'),
          'project': self.project.id
        })
        if (len(batch) >= batch_size):
          self.project.create_bulk_relations(batch, dedup=True)
          batch = []

    # Insert final batch
    if (len(batch) > 0):
      self.project.create_bulk_relations(batch, dedup=True)


  def create_source(self, source: dict):
    self.source_map[source.get('sourceId')] = [source.get('sourceName'), source.get('directoryName')]
    try:
      if source['type'] == 'text':
        self.project.create_text_source(source['sourceName'], source['text'], source['directoryName'])
      elif source['type'] == 'pdf':
        pdf_filepath = os.path.join(self.unpack_target_dir, source['directoryName'], source['sourceName'])
        text_bounds = self.__find_source_bounds(source['sourceId'])
        if (text_bounds is None):
          logger.error(f'Unable to find text bounds for {source["sourceId"]}')

        self.project.create_pdf_source(
          pdf_filepath,
          source['sourceName'],
          source['directoryName'],
          ocr=False,
          sourceText=source['text'],
          textBounds=text_bounds['textBounds']
        )
    except HTTPError as e:
      if (e.response.status_code == HTTPStatus.CONFLICT):
        logger.warning(f'Source {source.get("directory")}/{source.get("sourceName")} already exists. Skipping')
      else:
        raise e


  def __find_entity_files(self):
    contents = os.listdir(self.unpack_target_dir)
    export_files = list(filter(lambda item: re.match('.*jsonl', item), contents))

    self.source_file = self.__find_file(export_files, '.*\.sources\.jsonl')
    self.bounds_file = self.__find_file(export_files, '.*\.text-bounds\.jsonl')
    self.annotations_file = self.__find_file(export_files, '.*\.annotations\.jsonl')
    self.layers_file = self.__find_file(export_files, '.*\.layers\.jsonl')
    self.relations_file = self.__find_file(export_files, '.*\.relations\.jsonl')
    self.atntypes_file = self.__find_file(export_files, '.*\.atntypes\.jsonl')

    if (self.source_file is None):
      raise Exception('Sources missing from export. Ensure to make the export request using includeSources=True.')
    if (self.bounds_file is None):
      raise Exception('Text Bounds missing from export. Ensure to make the export request using includeTextBounds=True.')
    if (self.atntypes_file is None):
      raise Exception('Annotation Types missing from export. Ensure to make the export request using includeAnnotationTypes=True.')
    if (self.annotations_file is None):
      raise Exception('Annotations missing from export.')
    if (self.layers_file is None):
      raise Exception('Layers missing from export.')
    if (self.relations_file is None):
      raise Exception('Relations missing from export.')


  def __find_file(self, file_list: List[str], pattern: str):
    for filename in file_list:
      if (re.match(pattern, filename) is not None):
        return filename

    return None


  def __find_source_bounds(self, source_id: int):
    """Finds the source bounds of a source by iterating through the bounds file for the source id.

      *** Note: This is a highly inefficient v0.1 implementation ***
      The bounds file is redundantly iterated through every time.
      Perhaps the file should be pre-parsed, storing the byte offset of every source id to form
      an index?
    """
    filepath = os.path.join(self.unpack_target_dir, self.bounds_file)
    with jsonlines.open(filepath) as text_bounds:
      for bounds in text_bounds:
        if (bounds['sourceId'] == source_id):
          return bounds

    return None


import functools

# https://stackoverflow.com/questions/20535342/lazy-evaluation-in-python
class cached_property(object):
  def __init__(self, function):
    self.function = function
    functools.update_wrapper(self, function)

  def __get__(self, obj, type_):
    if obj is None:
      return self
    val = self.function(obj)
    obj.__dict__[self.function.__name__] = val
    return val


import json
from logging import Logger
import requests
import shutil

from annolab.api_helper import ApiHelper
from annolab import endpoints
from polling2 import poll

from enum import Enum

logger = Logger(__name__)

class ExportStatus(Enum):
  initialized = 'initalized'
  started     = 'started'
  errored     = 'errored'
  finished    = 'finished'

class ProjectExport:

  # How often to poll for export status, in seconds.
  poll_rate = 5

  def __init__(
    self,
    api_helper: ApiHelper,
    project,
    options: dict
  ):
    self.__api = api_helper
    self.project = project
    self.options = options
    self.status_url = None
    self.last_status = None
    self.download_url = None
    self.error = None


  def download_on_finish(self, filepath: str, timeout=3600):
    if (self.status_url is None):
      self.start()

    def check_completion():
      self.refresh_status()
      return self.last_status in [ExportStatus.finished.value, ExportStatus.errored.value]

    poll(
      lambda: check_completion(),
      step=self.poll_rate,
      timeout=timeout
    )

    with requests.get(self.download_url, stream=True) as r:
      with open(filepath, 'wb') as f:
        shutil.copyfileobj(r.raw, f)


  def start(self):
    body = {
      'projectIdentifier': self.project.name,
      'groupName' : self.project.owner_name,
      'includeAnnotationTypes': self.options.get('include_annotation_types', False),
      'includeSources': self.options.get('include_sources', False),
      'includeTextBounds': self.options.get('include_text_bounds', False)
    }

    if (self.options.get('source_ids', None) is not None):
      body['sourceIds'] = self.options.source_ids
    if (self.options.get('layers', None) is not None):
      body['annotationLayerNames'] = self.options.layers

    res = self.__api.post_request(
      endpoints.Export.post_export_project(),
      body
    )

    self.status_url = res.json().get('exportStatusUrl', None)

    if (self.status_url is None):
      raise Exception(f'Export status url not returned with response: {json.dumps(res)}')


  def refresh_status(self):
    if (self.status_url is None):
      raise Exception(f'Unable to request export status. No status url. Did you start the export using export.start()?')

    res = self.__api.get_request(self.status_url)

    body = res.json()
    self.last_status = body.get('status', self.last_status)

    if (self.last_status == ExportStatus.finished.value):
      self.download_url = body.get('downloadUrl')

    if (self.last_status == ExportStatus.errored.value):
      self.error = body.get('error')

    return self.last_status




from typing import Dict


class AnnotationRelation:

  @staticmethod
  def create_api_relation(dict: Dict):
    """
    Maps an sdk relation dict to an api annotation relation dict, for use in calls to the api.

    AnnotationRelation parameters:
      annotations:      [Union[str, int], Union[str, int]]
      type:             str
      value:            str
      reviewed          bool (Optional)
      project           Union[str, int]
    """
    relation = {
      'predecessorId': str(dict['annotations'][0]),
      'successorId': str(dict['annotations'][1])
    }

    if ('type' in dict): relation['annoTypeIdentifier'] = dict['type']
    if ('value' in dict): relation['value'] = dict['value']
    if ('reviewed' in dict): relation['isReviewed'] = dict['reviewed']
    if ('project' in dict): relation['projectIdentifier'] = dict['project']

    return relation


import io
from os import path
from typing import Any, Dict, List, Union
import requests

from annolab import endpoints
from annolab.api_helper import ApiHelper
from annolab.annotation import Annotation
from annolab.annotation_relation import AnnotationRelation
from annolab.project_import import ProjectImport
from annolab.project_export import ProjectExport

class Project:

  def __init__(
    self,
    name: str,
    id: int,
    owner_name: str,
    owner_id: int,
    default_dir: str,
    api_helper: ApiHelper
  ):
    self.name = name
    self.id = id
    self.owner_name = owner_name
    self.owner_id = owner_id
    self.default_dir = default_dir
    self.__api = api_helper


  @property
  def project_path(self):
    return f'{self.owner_name}/{self.name}'


  def find_source(self, name: str, directory: str = None):
    """
      Search for a source within a project by name and (optionally) directory.
      If directory is not provided, the default directory is used (typically "Uploads").
    """
    res = self.__api.get_request(
      endpoints.Source.get_source_by_path(
        owner_name=self.__api.default_owner['groupName'],
        project_name=self.name,
        directory_name=directory or self.default_dir,
        source_ref_name=name
      )
    )

    return res.json()


  def create_text_source(self, name: str, text: str, directory: str = None):
    """
      Creates a text source.
      If directory is not provided, the default directory is used (typically "Uploads").
    """
    body = {
      'projectIdentifier': self.id or self.name,
      'groupName': self.owner_name,
      'sourceName': name,
      'text': text
    }

    if (directory is not None):
      body['directoryIdentifier'] = directory

    res = self.__api.post_request(
      endpoints.Source.post_create_text(),
      body
    )

    return res.json()


  def create_pdf_source(
    self, 
    file: Union[str, io.IOBase, bytes], 
    name: str = None,
    directory: str = None,
    ocr: bool = False,
    preprocessor: str = 'none',
    timeout: float = 30.0,
    **params: dict):
    """
      Creates a pdf source from a local file, bytes, or filelike object.
      If directory is not provided, the default directory is used (typically "Uploads").
      Will OCR using your account preferred OCR if ocr parameter is set to True (Org only)
    """
    is_io_or_bytes = isinstance(file, io.IOBase) or isinstance(file, bytes)
    if (is_io_or_bytes and name is None):
      raise Exception('You must provide a name when passing a filelike object for pdf source creation')

    name = name or path.basename(file)

    init_res = self.__api.post_request(
      endpoints.Source.post_initialize_pdf(),
      {
        'projectIdentifier': self.id or self.name,
        'groupName': self.owner_name,
        'directoryIdentifier': directory or self.default_dir,
        'sourceName': name,
      },
      timeout=timeout
    )

    upload_url = init_res.json()['uploadUrl']

    pdf_file = file if is_io_or_bytes else open(file, 'r+b')
    self.__api.put_request(upload_url, data = pdf_file, headers={'Content-Type': 'application/pdf'})

    if (not isinstance(file, bytes)):
      pdf_file.close()

    body = {
      'projectIdentifier': self.id or self.name,
      'groupName': self.owner_name,
      'directoryIdentifier': directory or self.default_dir,
      'sourceIdentifier': name,
      'ocr': ocr,
      'preprocessor': preprocessor
    }

    body.update(params)

    create_res = self.__api.post_request(
      endpoints.Source.post_create_pdf(),
      body,
      timeout=timeout
    )

    return create_res.json()


  def create_pdf_source_from_web(self, url: str, name: str = None, directory: str = None, **params: dict):
    """
      Creates a pdf source from a web url.
      If directory is not provided, the default directory is used (typically "Uploads").
    """
    name = name or path.basename(url)
    res = requests.get(url)

    if (res.status_code >= 300):
      res.raise_for_status()

    return self.create_pdf_source(res.content, name, directory, params)


  def create_annotations(
    self,
    source_name: str,
    annotations: List[Any],
    relations: List[Any] = [],
    dedup: bool = True,
    directory: str = None):
    """
      Create annotations against a single source.

      Annotation parameters:
        type:       str  (Required)
        client_id:  str  (Optional, Required if passing relations)
        value:      str  (Optional)
        offsets:    [int, int] (Optional)
        text_bounds dict { 'type': 'Polygon' or 'MultiPolygon', coordinates: List[][][] } (Optional)
        image_bounds dict { 'type': 'Polygon' or 'MultiPolygon', coordinates: List[][][] } (Optional)
        bbox:       [int, int, int, int] (Optional)
        layer:      str  (Optional)
        page:       int  (Optional) Required for classification annotations.
        end_page    int  (Optional) Required for classification annotations.
        reviewed    bool (Optional)
    """
    directory = directory or self.default_dir

    res = self.__api.post_request(
      endpoints.Source.post_annotations(self.owner_name, self.name, directory, source_name),
      {
        'annotations': list(map(Annotation.create_api_annotation, annotations)),
        'relations': list(map(AnnotationRelation.create_api_relation, relations)),
        'preventDuplication': dedup
      })

    return res.json()


  def create_bulk_annotations(
    self,
    annotations: List[Any],
    dedup = True,
  ):
    """
      Create bulk annotations against one or more sources.
      To use this insert method, the annotation must include the project name or id.

      Annotation parameters:
        source     str or int (Required)
        project    str or int (Required)
        type:      str  (Required)
        client_id: str  (Optional, Required if passing relations)
        value:     str  (Optional)
        offsets:   [int, int] (Optional)
        text_bounds dict { 'type': 'Polygon' or 'MultiPolygon', coordinates: List[][][] } (Optional)
        image_bounds dict { 'type': 'Polygon' or 'MultiPolygon', coordinates: List[][][] } (Optional)
        bbox:      [int, int, int, int] (Optional)
        layer:     str  (Optional)
        page:      int  (Optional) Required for classification annotations.
        end_page   int  (Optional) Required for classification annotations.
        reviewed   bool (Optional)
    """
    res = self.__api.post_request(
      endpoints.Annotation.post_bulk_create(),
      {
        'annotations': list(map(Annotation.create_api_annotation, annotations)),
        'preventDuplication': dedup
      }
    )

    return res.json()


  def create_bulk_relations(
    self,
    relations: List[Any],
    dedup = True
  ):
    """
    Create bulk relations against one or more sources.
    To use this insert method, the annotation must include the project name or id.

    AnnotationRelation parameters:
      annotations:      [Union[str, int], Union[str, int]]
      type:             str  (Required)
      value:            str  (Optional)
      reviewed          bool (Optional)
      project           Union[str, int]
    """
    res = self.__api.post_request(
      endpoints.AnnotationRelation.post_bulk_create(),
      {
        'relations': list(map(AnnotationRelation.create_api_relation, relations)),
        'preventDuplication': dedup
      }
    )

    return res.json()


  def create_annotation_type(self, name: str, **kargs):
    """
      Create an annotation type.

      Annotation type parameters:
        name:        str  (Required)
        color:       str  (Optional)
        category:    str  (Optional)
        is_relation: str  (Optional, Defaults to false)
        is_document_classification: str  (Optional, Defaults to false)
    """
    res = self.__api.post_request(
      endpoints.AnnotationType.post_create(),
      {
        'category': kargs.get('category'),
        'projectIdentifier': self.id,
        'typeName': name,
        'color': kargs.get('color', None),
        'isRelation': kargs.get('is_relation', False),
        'isDocumentClassification': kargs.get('is_document_classification', False),
      }
    )

    return res.json()


  def create_annotation_layer(self, name: str, is_gold: bool = False, description: str = None):
    """
      Create an annotation layer.

      Annotation type parameters:
        name:        str  (Required)
        is_gold:     bool (Optional, Defaults to False)
        description: str  (Optional)
    """
    res = self.__api.post_request(
      endpoints.AnnotationLayer.post_create(),
      {
        'projectIdentifier': self.id,
        'layerName': name,
        'isGold': is_gold,
        'description': description,
      }
    )

    return res.json()


  def export(
    self,
    filepath: str,
    source_ids: List[int] = None,
    layers: List[str] = None,
    include_annotation_types: bool = False,
    include_sources: bool = False,
    include_text_bounds: bool = False,
    timeout: int = 3600
  ):
    body = {
      'projectIdentifier': self.name,
      'groupName' : self.owner_name,
      'includeAnnotationTypes': include_annotation_types,
      'includeSources': include_sources,
      'includeTextBounds': include_text_bounds
    }

    if (source_ids is not None): body['sourceIds'] = source_ids
    if (layers is not None): body['annotationLayerNames'] = layers

    export = ProjectExport(
      self.__api,
      self,
      {
        'source_ids': source_ids,
        'layers': layers,
        'include_annotation_types': include_annotation_types,
        'include_sources': include_sources,
        'include_text_bounds': include_text_bounds
      })

    export.start()
    export.download_on_finish(filepath, timeout=timeout)


  def update_from_export(self, filepath: str, skip_sources=False):
    project_import = ProjectImport(filepath, self, self.owner_name)

    project_import.unzip_export()

    if (skip_sources):
      project_import.import_annotation_types()
      project_import.import_layers()
      project_import.create_source_map()
      project_import.import_annotations()
      project_import.import_relations()
    else:
      project_import.import_all()

    project_import.cleanup()


  @staticmethod
  def create_from_response_json(resp_json: Dict, api_helper: ApiHelper):
    return Project(
      resp_json['name'],
      resp_json['id'],
      resp_json['groupName'],
      resp_json['groupId'],
      resp_json['defaultDirectory'],
      api_helper=api_helper,)



from typing import Dict

class Annotation:

  @staticmethod
  def create_api_annotation(dict: Dict):
    """
    Maps an sdk annotation dict to an api annotation dict, for use in calls to the api.

    Annotation parameters:
      type:       str  (Required)
      client_id:  str  (Optional, Required if passing relations)
      value:      str  (Optional)
      offsets:    [int, int] (Optional)
      text_bounds { 'type': 'Polygon', coordinates: List[List[List[int]]] } (Optional)
      bbox:       [int, int, int, int] (Optional, Ignored if text_bounds are passed)
      layer:      str  (Optional)
      page:       int  (Optional)
      reviewed    bool (Optional)
    """
    annotation = { 'annoTypeIdentifier': dict['type'] }

    if ('client_id' in dict): annotation['clientId'] = str(dict['client_id'])
    if ('offsets' in dict): annotation['offsets'] = dict['offsets']
    if ('value' in dict): annotation['value'] = dict['value']
    if ('bbox' in dict): annotation['bbox'] = dict['bbox']
    if ('text_bounds' in dict): annotation['textBounds'] = dict['text_bounds']
    if ('image_bounds' in dict): annotation['imageBounds'] = dict['image_bounds']
    if ('layer' in dict): annotation['layerIdentifier'] = dict['layer']
    if ('page' in dict): annotation['pageNumber'] = dict['page']
    if ('endPage' in dict): annotation['endPageNumber'] = dict['endPage']
    if ('reviewed' in dict): annotation['isReviewed'] = dict['reviewed']
    if ('source' in dict): annotation['sourceIdentifier'] = dict['source']
    if ('directory' in dict): annotation['directoryIdentifier'] = dict['directory']
    if ('project' in dict): annotation['projectIdentifier'] = dict['project']

    return annotation